import sys
from pathlib import Path
from unittest.mock import MagicMock

import pytest
from prefect import flow
from prefect.utilities.filesystem import relative_path_to_current_platform
from pydantic import SecretBytes, SecretStr

from prefect_snowflake.credentials import InvalidPemFormat, SnowflakeCredentials
from prefect_snowflake.database import SnowflakeConnector


def test_snowflake_credentials_init(credentials_params):
    snowflake_credentials = SnowflakeCredentials(**credentials_params)
    actual_credentials_params = snowflake_credentials.dict()
    for param in credentials_params:
        actual = actual_credentials_params[param]
        expected = credentials_params[param]
        if isinstance(actual, SecretStr):
            actual = actual.get_secret_value()
        assert actual == expected


def test_snowflake_credentials_validate_auth_kwargs(credentials_params):
    credentials_params_missing = credentials_params.copy()
    credentials_params_missing.pop("password")
    with pytest.raises(ValueError, match="One of the authentication keys"):
        SnowflakeCredentials(**credentials_params_missing)


def test_snowflake_credentials_validate_auth_kwargs_private_keys(credentials_params):
    credentials_params["private_key"] = "key"
    credentials_params["private_key_path"] = "keypath"
    with pytest.raises(ValueError, match="Do not provide both private_key and private"):
        SnowflakeCredentials(**credentials_params)


def test_snowflake_credentials_validate_auth_kwargs_private_key_password(
    credentials_params,
):
    credentials_params["private_key_passphrase"] = "key"
    with pytest.raises(ValueError, match="Do not provide both password and private_"):
        SnowflakeCredentials(**credentials_params)


def test_snowflake_credentials_validate_token_kwargs(credentials_params):
    credentials_params_missing = credentials_params.copy()
    credentials_params_missing.pop("password")
    credentials_params_missing["authenticator"] = "oauth"
    with pytest.raises(ValueError, match="If authenticator is set to `oauth`"):
        SnowflakeCredentials(**credentials_params_missing)

    # now test if passing both works
    credentials_params_missing["token"] = "some_token"
    assert SnowflakeCredentials(**credentials_params_missing)


def test_snowflake_credentials_validate_okta_endpoint_kwargs(credentials_params):
    credentials_params_missing = credentials_params.copy()
    credentials_params_missing.pop("password")
    credentials_params_missing["authenticator"] = "okta_endpoint"
    with pytest.raises(ValueError, match="If authenticator is set to `okta_endpoint`"):
        SnowflakeCredentials(**credentials_params_missing)

    # now test if passing both works
    credentials_params_missing["endpoint"] = "https://account_name.okta.com"
    snowflake_credentials = SnowflakeCredentials(**credentials_params_missing)
    assert snowflake_credentials.endpoint == "https://account_name.okta.com"


def test_snowflake_credentials_support_deprecated_okta_endpoint(credentials_params):
    credentials_params_missing = credentials_params.copy()
    credentials_params_missing.pop("password")
    credentials_params_missing["authenticator"] = "okta_endpoint"
    credentials_params_missing["okta_endpoint"] = "deprecated.com"
    snowflake_credentials = SnowflakeCredentials(**credentials_params_missing)
    assert snowflake_credentials.endpoint == "deprecated.com"


def test_snowflake_credentials_support_endpoint_overrides_okta_endpoint(
    credentials_params,
):
    credentials_params_missing = credentials_params.copy()
    credentials_params_missing.pop("password")
    credentials_params_missing["authenticator"] = "okta_endpoint"
    credentials_params_missing["okta_endpoint"] = "deprecated.com"
    credentials_params_missing["endpoint"] = "new.com"
    snowflake_credentials = SnowflakeCredentials(**credentials_params_missing)
    assert snowflake_credentials.endpoint == "new.com"


def test_snowflake_private_credentials_init(private_credentials_params):
    snowflake_credentials = SnowflakeCredentials(**private_credentials_params)
    actual_credentials_params = snowflake_credentials.dict()
    for param in private_credentials_params:
        actual = actual_credentials_params[param]
        expected = private_credentials_params[param]
        if isinstance(actual, (SecretStr, SecretBytes)):
            actual = actual.get_secret_value()
        if sys.platform != "win32":
            assert actual == expected


def test_snowflake_private_credentials_malformed_certificate(
    private_credentials_params, private_malformed_credentials_params
):
    correct = SnowflakeCredentials(**private_credentials_params)
    malformed = SnowflakeCredentials(**private_malformed_credentials_params)
    c1 = correct.resolve_private_key()
    c2 = malformed.resolve_private_key()
    assert isinstance(c1, bytes)
    assert isinstance(c2, bytes)
    assert c1 == c2


def test_snowflake_private_credentials_invalid_certificate(private_credentials_params):
    private_credentials_params["private_key"] = "---- INVALID CERTIFICATE ----"
    with pytest.raises(InvalidPemFormat):
        SnowflakeCredentials(**private_credentials_params).resolve_private_key()


def test_snowflake_credentials_validate_private_key_password(
    private_credentials_params,
):
    credentials_params_missing = private_credentials_params.copy()
    password = credentials_params_missing.pop("password")
    private_key = credentials_params_missing.pop("private_key")
    assert password == "letmein"
    assert isinstance(private_key, bytes)
    # Test cert as string
    credentials = SnowflakeCredentials(**private_credentials_params)
    assert credentials.resolve_private_key() is not None


def test_snowflake_credentials_validate_private_key_passphrase(
    private_credentials_params,
):
    private_credentials_params[
        "private_key_passphrase"
    ] = private_credentials_params.pop("password")
    credentials_params_missing = private_credentials_params.copy()
    password = credentials_params_missing.pop("private_key_passphrase")
    private_key = credentials_params_missing.pop("private_key")
    assert password == "letmein"
    assert isinstance(private_key, bytes)
    # Test cert as string
    credentials = SnowflakeCredentials(**private_credentials_params)
    assert credentials.resolve_private_key() is not None


def test_snowflake_credentials_validate_private_key_path(
    private_credentials_params, tmp_path
):
    private_key_path = tmp_path / "private_key.pem"
    private_key_path.write_bytes(private_credentials_params.pop("private_key"))
    private_credentials_params["private_key_path"] = private_key_path
    private_credentials_params[
        "private_key_passphrase"
    ] = private_credentials_params.pop("password")
    credentials = SnowflakeCredentials(**private_credentials_params)
    assert credentials.resolve_private_key() is not None


def test_snowflake_credentials_validate_private_key_invalid(private_credentials_params):
    credentials_params_missing = private_credentials_params.copy()
    private_key = credentials_params_missing.pop("private_key")
    assert isinstance(private_key, bytes)
    assert private_key.startswith(b"----")
    with pytest.raises(ValueError, match="Bad decrypt. Incorrect password?"):
        credentials = SnowflakeCredentials(**private_credentials_params)
        credentials.password = "_wrong_password"
        assert credentials.resolve_private_key() is not None


def test_snowflake_credentials_validate_private_key_unexpected_password(
    private_credentials_params,
):
    credentials_params_missing = private_credentials_params.copy()
    private_key = credentials_params_missing.pop("private_key")
    assert isinstance(private_key, bytes)
    assert private_key.startswith(b"----")
    with pytest.raises(
        TypeError, match="Password was not given but private key is encrypted"
    ):
        credentials = SnowflakeCredentials(**private_credentials_params)
        credentials.password = None
        assert credentials.resolve_private_key() is not None


def test_snowflake_credentials_validate_private_key_no_pass_password(
    private_no_pass_credentials_params,
):
    credentials_params_missing = private_no_pass_credentials_params.copy()
    password = credentials_params_missing.pop("password")
    private_key = credentials_params_missing.pop("private_key")
    assert password == "letmein"
    assert isinstance(private_key, bytes)
    assert private_key.startswith(b"----")

    with pytest.raises(
        TypeError, match="Password was given but private key is not encrypted"
    ):
        credentials = SnowflakeCredentials(**private_no_pass_credentials_params)
        assert credentials.resolve_private_key() is not None


def test_snowflake_credentials_validate_private_key_is_pem(
    private_no_pass_credentials_params,
):
    private_no_pass_credentials_params["private_key"] = "_invalid_key_"
    with pytest.raises(InvalidPemFormat):
        credentials = SnowflakeCredentials(**private_no_pass_credentials_params)
        assert credentials.resolve_private_key() is not None


def test_snowflake_credentials_validate_private_key_is_pem_bytes(
    private_no_pass_credentials_params,
):
    private_no_pass_credentials_params["private_key"] = "_invalid_key_"
    with pytest.raises(InvalidPemFormat):
        credentials = SnowflakeCredentials(**private_no_pass_credentials_params)
        assert credentials.resolve_private_key() is not None


def test_snowflake_credentials_validate_private_key_path_init(
    private_key_path_credentials_params,
):
    snowflake_credentials = SnowflakeCredentials(**private_key_path_credentials_params)
    actual_credentials_params = snowflake_credentials.dict()
    for param in private_key_path_credentials_params:
        actual = actual_credentials_params[param]
        expected = private_key_path_credentials_params[param]
        if isinstance(actual, (SecretStr, SecretBytes)):
            actual = actual.get_secret_value()
        elif isinstance(actual, Path):
            actual = relative_path_to_current_platform(actual)
            expected = relative_path_to_current_platform(expected)
        assert actual == expected


def test_get_client(credentials_params, snowflake_connect_mock: MagicMock):
    snowflake_credentials = SnowflakeCredentials(**credentials_params)
    snowflake_credentials.get_client()
    snowflake_connect_mock.assert_called_with(
        application="Prefect_Snowflake_Collection",
        account="account",
        user="user",
        password="password",
    )


def test_get_client_okta_endpoint(
    credentials_params, snowflake_connect_mock: MagicMock
):
    okta_endpoint = "https://account_name.okta.com"
    credentials_params_okta_endpoint = credentials_params.copy()
    del credentials_params_okta_endpoint["password"]
    credentials_params_okta_endpoint["authenticator"] = "okta_endpoint"
    credentials_params_okta_endpoint["endpoint"] = okta_endpoint
    snowflake_credentials = SnowflakeCredentials(**credentials_params_okta_endpoint)
    snowflake_credentials.get_client()
    snowflake_connect_mock.assert_called_with(
        application="Prefect_Snowflake_Collection",
        account="account",
        user="user",
        authenticator="https://account_name.okta.com",
    )


def test_snowflake_credentials_deprecated_okta_endpoint(
    credentials_params, snowflake_connect_mock: MagicMock
):
    okta_endpoint = "https://account_name.okta.com"
    credentials_params_okta_endpoint = credentials_params.copy()
    del credentials_params_okta_endpoint["password"]
    credentials_params_okta_endpoint["authenticator"] = "okta_endpoint"
    credentials_params_okta_endpoint["endpoint"] = okta_endpoint
    snowflake_credentials = SnowflakeCredentials(**credentials_params_okta_endpoint)
    snowflake_credentials.get_client()
    snowflake_connect_mock.assert_called_with(
        application="Prefect_Snowflake_Collection",
        account="account",
        user="user",
        authenticator="https://account_name.okta.com",
    )


def test_snowflake_credentials_unencrypted_private_key_password(
    private_no_pass_credentials_params,
):
    snowflake_credentials = SnowflakeCredentials(**private_no_pass_credentials_params)
    assert snowflake_credentials.private_key is not None
    assert snowflake_credentials.password is not None
    # Raises error if invalid
    with pytest.raises(
        TypeError, match="Password was given but private key is not encrypted"
    ):
        snowflake_credentials.get_client()


def test_snowflake_credentials_unencrypted_private_key_no_password(
    private_no_pass_credentials_params, snowflake_connect_mock: MagicMock
):
    snowflake_credentials = SnowflakeCredentials(**private_no_pass_credentials_params)
    snowflake_credentials.password = None
    assert snowflake_credentials.private_key is not None
    # Raises error if invalid
    snowflake_credentials.get_client()


def test_snowflake_credentials_unencrypted_private_key_empty_password(
    private_no_pass_credentials_params, snowflake_connect_mock: MagicMock
):
    snowflake_credentials = SnowflakeCredentials(**private_no_pass_credentials_params)
    assert snowflake_credentials.private_key is not None

    snowflake_credentials.password = SecretBytes(b" ")
    snowflake_credentials.get_client()
    snowflake_credentials.password = SecretBytes(b"")
    snowflake_credentials.get_client()
    snowflake_credentials.password = SecretStr("")
    snowflake_credentials.get_client()
    snowflake_credentials.password = SecretStr("   ")
    snowflake_credentials.get_client()


def test_snowflake_credentials_encrypted_private_key_is_valid(
    private_credentials_params, snowflake_connect_mock: MagicMock
):
    snowflake_credentials = SnowflakeCredentials(**private_credentials_params)
    assert snowflake_credentials.private_key is not None
    assert snowflake_credentials.password is not None
    # Raises error if invalid
    snowflake_credentials.get_client()


def test_snowflake_with_no_private_key_flow():
    """
    https://github.com/PrefectHQ/prefect-snowflake/issues/62
    """

    @flow
    def snowflake_query_flow():
        snowflake_credentials = SnowflakeCredentials(
            account="account", user="user", password="password", role="MY_ROLE"
        )
        snowflake_connector = SnowflakeConnector(
            database="database",
            warehouse="warehouse",
            schema="schema",
            credentials=snowflake_credentials,
        )
        return snowflake_connector

    snowflake_connector = snowflake_query_flow()
    assert isinstance(snowflake_connector, SnowflakeConnector)
    assert isinstance(snowflake_connector.credentials, SnowflakeCredentials)
    assert snowflake_connector.credentials.private_key is None
    assert snowflake_connector.credentials.private_key_path is None
    assert snowflake_connector.credentials.private_key_passphrase is None
    assert snowflake_connector.credentials.password.get_secret_value() == "password"


def test_snowflake_with_private_key_path_flow():
    """
    https://github.com/PrefectHQ/prefect-snowflake/issues/62
    """

    @flow
    def snowflake_query_flow():
        snowflake_credentials = SnowflakeCredentials(
            account="account",
            user="user",
            private_key_path="private_key_path",
            private_key_passphrase="passphrase",
        )
        snowflake_connector = SnowflakeConnector(
            database="database",
            warehouse="warehouse",
            schema="schema",
            credentials=snowflake_credentials,
        )
        return snowflake_connector

    snowflake_connector = snowflake_query_flow()
    assert isinstance(snowflake_connector, SnowflakeConnector)
    assert isinstance(snowflake_connector.credentials, SnowflakeCredentials)
    assert snowflake_connector.credentials.private_key is None
    assert snowflake_connector.credentials.private_key_path == Path("private_key_path")
    assert (
        snowflake_connector.credentials.private_key_passphrase.get_secret_value()
        == "passphrase"
    )
    assert snowflake_connector.credentials.password is None



from . import _version
from prefect_snowflake.credentials import SnowflakeCredentials  # noqa
from prefect_snowflake.database import SnowflakeConnector  # noqa

__version__ = _version.get_versions()["version"]



from unittest.mock import MagicMock

import pytest
from prefect import flow
from pydantic import SecretBytes, SecretStr

from prefect_snowflake.database import (
    BEGIN_TRANSACTION_STATEMENT,
    END_TRANSACTION_STATEMENT,
    SnowflakeConnector,
    snowflake_multiquery,
    snowflake_query,
    snowflake_query_sync,
)


def test_snowflake_connector_init(connector_params):
    snowflake_connector = SnowflakeConnector(**connector_params)
    actual_connector_params = snowflake_connector.dict()
    for param in connector_params:
        expected = connector_params[param]
        if param == "schema":
            param = "schema_"
        actual = actual_connector_params[param]
        if isinstance(actual, SecretStr):
            actual = actual.get_secret_value()
        assert actual == expected


def test_snowflake_connector_password_is_secret_str(connector_params):
    snowflake_connector = SnowflakeConnector(**connector_params)
    password = snowflake_connector.credentials.password
    assert isinstance(password, SecretStr)
    assert password.get_secret_value() == "password"


def test_snowflake_connector_private_key_is_secret(private_connector_params):
    snowflake_connector = SnowflakeConnector(**private_connector_params)
    private_key = snowflake_connector.credentials.private_key
    assert isinstance(private_key, (SecretStr, SecretBytes))


class SnowflakeCursor:
    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False

    def execute_async(self, query, params):
        query_id = "1234"
        self.result = {query_id: [(query, params)]}
        return {"queryId": query_id}

    def get_results_from_sfqid(self, query_id):
        self.query_result = self.result[query_id]

    def fetchall(self):
        return self.query_result

    def execute(self, query, params=None):
        self.query_result = [(query, params, "sync")]
        return self


class SnowflakeConnection:
    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False

    def cursor(self, cursor_type):
        return SnowflakeCursor()

    def is_still_running(self, state):
        return state

    def get_query_status_throw_if_error(self, query_id):
        return False


@pytest.fixture()
def snowflake_connector(snowflake_connect_mock):
    snowflake_connector_mock = MagicMock()
    snowflake_connector_mock.get_connection.return_value = SnowflakeConnection()
    return snowflake_connector_mock


def test_snowflake_query(snowflake_connector):
    @flow
    def test_flow():
        result = snowflake_query(
            "query",
            snowflake_connector,
            params=("param",),
        )
        return result

    result = test_flow()
    assert result[0][0] == "query"
    assert result[0][1] == ("param",)


def test_snowflake_multiquery(snowflake_connector):
    @flow
    def test_flow():
        result = snowflake_multiquery(
            ["query1", "query2"],
            snowflake_connector,
            params=("param",),
        )
        return result

    result = test_flow()
    assert result[0][0][0] == "query1"
    assert result[0][0][1] == ("param",)
    assert result[1][0][0] == "query2"
    assert result[1][0][1] == ("param",)


def test_snowflake_multiquery_transaction(snowflake_connector):
    @flow
    def test_flow():
        result = snowflake_multiquery(
            ["query1", "query2"],
            snowflake_connector,
            params=("param",),
            as_transaction=True,
        )
        return result

    result = test_flow()
    assert result[0][0][0] == "query1"
    assert result[0][0][1] == ("param",)
    assert result[1][0][0] == "query2"
    assert result[1][0][1] == ("param",)


def test_snowflake_multiquery_transaction_with_transaction_control_results(
    snowflake_connector,
):
    @flow
    def test_flow():
        result = snowflake_multiquery(
            ["query1", "query2"],
            snowflake_connector,
            params=("param",),
            as_transaction=True,
            return_transaction_control_results=True,
        )
        return result

    result = test_flow()
    assert result[0][0][0] == BEGIN_TRANSACTION_STATEMENT
    assert result[1][0][0] == "query1"
    assert result[1][0][1] == ("param",)
    assert result[2][0][0] == "query2"
    assert result[2][0][1] == ("param",)
    assert result[3][0][0] == END_TRANSACTION_STATEMENT


def test_snowflake_query_sync(snowflake_connector):
    @flow()
    def test_snowflake_query_sync_flow():
        result = snowflake_query_sync("query", snowflake_connector, params=("param",))
        return result

    result = test_snowflake_query_sync_flow()
    assert result[0][0] == "query"
    assert result[0][1] == ("param",)
    assert result[0][2] == "sync"


def test_snowflake_private_connector_init(private_connector_params):
    snowflake_connector = SnowflakeConnector(**private_connector_params)
    actual_connector_params = snowflake_connector.dict()
    for param in private_connector_params:
        expected = private_connector_params[param]
        if param == "schema":
            param = "schema_"
        actual = actual_connector_params[param]
        if isinstance(actual, (SecretStr, SecretBytes)):
            actual = actual.get_secret_value()
        assert actual == expected


class TestSnowflakeConnector:
    @pytest.fixture
    def snowflake_connector(self, connector_params, snowflake_connect_mock):
        connector = SnowflakeConnector(**connector_params)
        return connector

    def test_block_initialization(self, snowflake_connector):
        assert snowflake_connector._connection is None
        assert snowflake_connector._unique_cursors is None

    def test_get_connection(self, snowflake_connector: SnowflakeConnector, caplog):
        connection = snowflake_connector.get_connection()
        assert snowflake_connector._connection is connection
        assert caplog.records[0].msg == "Started a new connection to Snowflake."

    def test_reset_cursors(self, snowflake_connector: SnowflakeConnector, caplog):
        mock_cursor = MagicMock()
        snowflake_connector.reset_cursors()
        assert caplog.records[0].msg == "There were no cursors to reset."

        snowflake_connector._start_connection()
        snowflake_connector._unique_cursors["12345"] = mock_cursor
        snowflake_connector.reset_cursors()
        assert len(snowflake_connector._unique_cursors) == 0
        mock_cursor.close.assert_called_once()

    def test_fetch_one(self, snowflake_connector: SnowflakeConnector):
        result = snowflake_connector.fetch_one("query", parameters=("param",))
        assert result == (0,)
        result = snowflake_connector.fetch_one("query", parameters=("param",))
        assert result == (1,)

    def test_fetch_many(self, snowflake_connector: SnowflakeConnector):
        result = snowflake_connector.fetch_many("query", parameters=("param",), size=2)
        assert result == [(0,), (1,)]
        result = snowflake_connector.fetch_many("query", parameters=("param",))
        assert result == [(2,)]

    def test_fetch_all(self, snowflake_connector: SnowflakeConnector):
        result = snowflake_connector.fetch_all("query", parameters=("param",))
        assert result == [(0,), (1,), (2,), (3,), (4,)]

    def test_execute(self, snowflake_connector: SnowflakeConnector):
        assert snowflake_connector.execute("query", parameters=("param",)) is None

    def test_execute_Many(self, snowflake_connector: SnowflakeConnector):
        assert (
            snowflake_connector.execute_many("query", seq_of_parameters=[("param",)])
            is None
        )

    def test_close(self, snowflake_connector: SnowflakeConnector, caplog):
        assert snowflake_connector.close() is None
        assert caplog.records[0].msg == "There were no cursors to reset."
        assert caplog.records[1].msg == "There was no connection open to be closed."

        snowflake_connector._start_connection()
        assert snowflake_connector.close() is None
        assert snowflake_connector._connection is None
        assert snowflake_connector._unique_cursors == {}

    def test_context_management(self, snowflake_connector):
        with snowflake_connector:
            assert snowflake_connector._connection is None
            assert snowflake_connector._unique_cursors is None

        assert snowflake_connector._connection is None
        assert snowflake_connector._unique_cursors is None



"""Module for querying against Snowflake databases."""

import asyncio
from typing import Any, Dict, List, Optional, Tuple, Union

from prefect import task
from prefect.blocks.abstract import DatabaseBlock
from prefect.utilities.asyncutils import run_sync_in_worker_thread, sync_compatible
from prefect.utilities.hashing import hash_objects
from pydantic import Field
from snowflake.connector.connection import SnowflakeConnection
from snowflake.connector.cursor import SnowflakeCursor

from prefect_snowflake import SnowflakeCredentials

BEGIN_TRANSACTION_STATEMENT = "BEGIN TRANSACTION"
END_TRANSACTION_STATEMENT = "COMMIT"


class SnowflakeConnector(DatabaseBlock):

    """
    Block used to manage connections with Snowflake.

    Upon instantiating, a connection is created and maintained for the life of
    the object until the close method is called.

    It is recommended to use this block as a context manager, which will automatically
    close the engine and its connections when the context is exited.

    It is also recommended that this block is loaded and consumed within a single task
    or flow because if the block is passed across separate tasks and flows,
    the state of the block's connection and cursor will be lost.

    Args:
        credentials: The credentials to authenticate with Snowflake.
        database: The name of the default database to use.
        warehouse: The name of the default warehouse to use.
        schema: The name of the default schema to use;
            this attribute is accessible through `SnowflakeConnector(...).schema_`.
        fetch_size: The number of rows to fetch at a time.
        poll_frequency_s: The number of seconds before checking query.

    Examples:
        Load stored Snowflake connector as a context manager:
        ```python
        from prefect_snowflake.database import SnowflakeConnector

        snowflake_connector = SnowflakeConnector.load("BLOCK_NAME"):
        ```

        Insert data into database and fetch results.
        ```python
        from prefect_snowflake.database import SnowflakeConnector

        with SnowflakeConnector.load("BLOCK_NAME") as conn:
            conn.execute(
                "CREATE TABLE IF NOT EXISTS customers (name varchar, address varchar);"
            )
            conn.execute_many(
                "INSERT INTO customers (name, address) VALUES (%(name)s, %(address)s);",
                seq_of_parameters=[
                    {"name": "Ford", "address": "Highway 42"},
                    {"name": "Unknown", "address": "Space"},
                    {"name": "Me", "address": "Myway 88"},
                ],
            )
            results = conn.fetch_all(
                "SELECT * FROM customers WHERE address = %(address)s",
                parameters={"address": "Space"}
            )
            print(results)
        ```
    """  # noqa

    _block_type_name = "Snowflake Connector"
    _logo_url = "https://images.ctfassets.net/gm98wzqotmnx/2DxzAeTM9eHLDcRQx1FR34/f858a501cdff918d398b39365ec2150f/snowflake.png?h=250"  # noqa
    _documentation_url = "https://prefecthq.github.io/prefect-snowflake/database/#prefect_snowflake.database.SnowflakeConnector"  # noqa
    _description = "Perform data operations against a Snowflake database."

    credentials: SnowflakeCredentials = Field(
        default=..., description="The credentials to authenticate with Snowflake."
    )
    database: str = Field(
        default=..., description="The name of the default database to use."
    )
    warehouse: str = Field(
        default=..., description="The name of the default warehouse to use."
    )
    schema_: str = Field(
        default=...,
        alias="schema",
        description="The name of the default schema to use.",
    )
    fetch_size: int = Field(
        default=1, description="The default number of rows to fetch at a time."
    )
    poll_frequency_s: int = Field(
        default=1,
        title="Poll Frequency [seconds]",
        description=(
            "The number of seconds between checking query "
            "status for long running queries."
        ),
    )

    _connection: Optional[SnowflakeConnection] = None
    _unique_cursors: Dict[str, SnowflakeCursor] = None

    def get_connection(self, **connect_kwargs: Dict[str, Any]) -> SnowflakeConnection:
        """
        Returns an authenticated connection that can be
        used to query from Snowflake databases.

        Args:
            **connect_kwargs: Additional arguments to pass to
                `snowflake.connector.connect`.

        Returns:
            The authenticated SnowflakeConnection.

        Examples:
            ```python
            from prefect_snowflake.credentials import SnowflakeCredentials
            from prefect_snowflake.database import SnowflakeConnector

            snowflake_credentials = SnowflakeCredentials(
                account="account",
                user="user",
                password="password",
            )
            snowflake_connector = SnowflakeConnector(
                database="database",
                warehouse="warehouse",
                schema="schema",
                credentials=snowflake_credentials
            )
            with snowflake_connector.get_connection() as connection:
                ...
            ```
        """
        if self._connection is not None:
            return self._connection

        connect_params = {
            "database": self.database,
            "warehouse": self.warehouse,
            "schema": self.schema_,
        }
        connection = self.credentials.get_client(**connect_kwargs, **connect_params)
        self._connection = connection
        self.logger.info("Started a new connection to Snowflake.")
        return connection

    def _start_connection(self):
        """
        Starts Snowflake database connection.
        """
        self.get_connection()
        if self._unique_cursors is None:
            self._unique_cursors = {}

    def _get_cursor(self, inputs: Dict[str, Any]) -> Tuple[bool, SnowflakeCursor]:
        """
        Get a Snowflake cursor.

        Args:
            inputs: The inputs to generate a unique hash, used to decide
                whether a new cursor should be used.

        Returns:
            Whether a cursor is new and a Snowflake cursor.
        """
        self._start_connection()

        input_hash = hash_objects(inputs)
        if input_hash is None:
            raise RuntimeError(
                "We were not able to hash your inputs, "
                "which resulted in an unexpected data return; "
                "please open an issue with a reproducible example."
            )
        if input_hash not in self._unique_cursors.keys():
            new_cursor = self._connection.cursor()
            self._unique_cursors[input_hash] = new_cursor
            return True, new_cursor
        else:
            existing_cursor = self._unique_cursors[input_hash]
            return False, existing_cursor

    async def _execute_async(self, cursor: SnowflakeCursor, inputs: Dict[str, Any]):
        """Helper method to execute operations asynchronously."""
        response = await run_sync_in_worker_thread(cursor.execute_async, **inputs)
        self.logger.info(
            f"Executing the operation, {inputs['command']!r}, asynchronously; "
            f"polling for the result every {self.poll_frequency_s} seconds."
        )

        query_id = response["queryId"]
        while self._connection.is_still_running(
            await run_sync_in_worker_thread(
                self._connection.get_query_status_throw_if_error, query_id
            )
        ):
            await asyncio.sleep(self.poll_frequency_s)
        await run_sync_in_worker_thread(cursor.get_results_from_sfqid, query_id)

    def reset_cursors(self) -> None:
        """
        Tries to close all opened cursors.

        Examples:
            Reset the cursors to refresh cursor position.
            ```python
            from prefect_snowflake.database import SnowflakeConnector

            with SnowflakeConnector.load("BLOCK_NAME") as conn:
                conn.execute(
                    "CREATE TABLE IF NOT EXISTS customers (name varchar, address varchar);"
                )
                conn.execute_many(
                    "INSERT INTO customers (name, address) VALUES (%(name)s, %(address)s);",
                    seq_of_parameters=[
                        {"name": "Ford", "address": "Highway 42"},
                        {"name": "Unknown", "address": "Space"},
                        {"name": "Me", "address": "Myway 88"},
                    ],
                )
                print(conn.fetch_one("SELECT * FROM customers"))  # Ford
                conn.reset_cursors()
                print(conn.fetch_one("SELECT * FROM customers"))  # should be Ford again
            ```
        """  # noqa
        if not self._unique_cursors:
            self.logger.info("There were no cursors to reset.")
            return

        input_hashes = tuple(self._unique_cursors.keys())
        for input_hash in input_hashes:
            cursor = self._unique_cursors.pop(input_hash)
            try:
                cursor.close()
            except Exception as exc:
                self.logger.warning(
                    f"Failed to close cursor for input hash {input_hash!r}: {exc}"
                )
        self.logger.info("Successfully reset the cursors.")

    @sync_compatible
    async def fetch_one(
        self,
        operation: str,
        parameters: Optional[Dict[str, Any]] = None,
        **execute_kwargs: Dict[str, Any],
    ) -> Tuple[Any]:
        """
        Fetch a single result from the database.
        Repeated calls using the same inputs to *any* of the fetch methods of this
        block will skip executing the operation again, and instead,
        return the next set of results from the previous execution,
        until the reset_cursors method is called.

        Args:
            operation: The SQL query or other operation to be executed.
            parameters: The parameters for the operation.
            **execute_kwargs: Additional options to pass to `cursor.execute_async`.

        Returns:
            A tuple containing the data returned by the database,
                where each row is a tuple and each column is a value in the tuple.

        Examples:
            Fetch one row from the database where address is Space.
            ```python
            from prefect_snowflake.database import SnowflakeConnector

            with SnowflakeConnector.load("BLOCK_NAME") as conn:
                conn.execute(
                    "CREATE TABLE IF NOT EXISTS customers (name varchar, address varchar);"
                )
                conn.execute_many(
                    "INSERT INTO customers (name, address) VALUES (%(name)s, %(address)s);",
                    seq_of_parameters=[
                        {"name": "Ford", "address": "Highway 42"},
                        {"name": "Unknown", "address": "Space"},
                        {"name": "Me", "address": "Myway 88"},
                    ],
                )
                result = conn.fetch_one(
                    "SELECT * FROM customers WHERE address = %(address)s",
                    parameters={"address": "Space"}
                )
                print(result)
            ```
        """  # noqa
        inputs = dict(
            command=operation,
            params=parameters,
            **execute_kwargs,
        )
        new, cursor = self._get_cursor(inputs)
        if new:
            await self._execute_async(cursor, inputs)
        self.logger.debug("Preparing to fetch a row.")
        result = await run_sync_in_worker_thread(cursor.fetchone)
        return result

    @sync_compatible
    async def fetch_many(
        self,
        operation: str,
        parameters: Optional[Dict[str, Any]] = None,
        size: Optional[int] = None,
        **execute_kwargs: Dict[str, Any],
    ) -> List[Tuple[Any]]:
        """
        Fetch a limited number of results from the database.
        Repeated calls using the same inputs to *any* of the fetch methods of this
        block will skip executing the operation again, and instead,
        return the next set of results from the previous execution,
        until the reset_cursors method is called.

        Args:
            operation: The SQL query or other operation to be executed.
            parameters: The parameters for the operation.
            size: The number of results to return; if None or 0, uses the value of
                `fetch_size` configured on the block.
            **execute_kwargs: Additional options to pass to `cursor.execute_async`.

        Returns:
            A list of tuples containing the data returned by the database,
                where each row is a tuple and each column is a value in the tuple.

        Examples:
            Repeatedly fetch two rows from the database where address is Highway 42.
            ```python
            from prefect_snowflake.database import SnowflakeConnector

            with SnowflakeConnector.load("BLOCK_NAME") as conn:
                conn.execute(
                    "CREATE TABLE IF NOT EXISTS customers (name varchar, address varchar);"
                )
                conn.execute_many(
                    "INSERT INTO customers (name, address) VALUES (%(name)s, %(address)s);",
                    seq_of_parameters=[
                        {"name": "Marvin", "address": "Highway 42"},
                        {"name": "Ford", "address": "Highway 42"},
                        {"name": "Unknown", "address": "Highway 42"},
                        {"name": "Me", "address": "Highway 42"},
                    ],
                )
                result = conn.fetch_many(
                    "SELECT * FROM customers WHERE address = %(address)s",
                    parameters={"address": "Highway 42"},
                    size=2
                )
                print(result)  # Marvin, Ford
                result = conn.fetch_many(
                    "SELECT * FROM customers WHERE address = %(address)s",
                    parameters={"address": "Highway 42"},
                    size=2
                )
                print(result)  # Unknown, Me
            ```
        """  # noqa
        inputs = dict(
            command=operation,
            params=parameters,
            **execute_kwargs,
        )
        new, cursor = self._get_cursor(inputs)
        if new:
            await self._execute_async(cursor, inputs)
        size = size or self.fetch_size
        self.logger.debug(f"Preparing to fetch {size} rows.")
        result = await run_sync_in_worker_thread(cursor.fetchmany, size=size)
        return result

    @sync_compatible
    async def fetch_all(
        self,
        operation: str,
        parameters: Optional[Dict[str, Any]] = None,
        **execute_kwargs: Dict[str, Any],
    ) -> List[Tuple[Any]]:
        """
        Fetch all results from the database.
        Repeated calls using the same inputs to *any* of the fetch methods of this
        block will skip executing the operation again, and instead,
        return the next set of results from the previous execution,
        until the reset_cursors method is called.

        Args:
            operation: The SQL query or other operation to be executed.
            parameters: The parameters for the operation.
            **execute_kwargs: Additional options to pass to `cursor.execute_async`.

        Returns:
            A list of tuples containing the data returned by the database,
                where each row is a tuple and each column is a value in the tuple.

        Examples:
            Fetch all rows from the database where address is Highway 42.
            ```python
            from prefect_snowflake.database import SnowflakeConnector

            with SnowflakeConnector.load("BLOCK_NAME") as conn:
                conn.execute(
                    "CREATE TABLE IF NOT EXISTS customers (name varchar, address varchar);"
                )
                conn.execute_many(
                    "INSERT INTO customers (name, address) VALUES (%(name)s, %(address)s);",
                    seq_of_parameters=[
                        {"name": "Marvin", "address": "Highway 42"},
                        {"name": "Ford", "address": "Highway 42"},
                        {"name": "Unknown", "address": "Highway 42"},
                        {"name": "Me", "address": "Myway 88"},
                    ],
                )
                result = conn.fetch_all(
                    "SELECT * FROM customers WHERE address = %(address)s",
                    parameters={"address": "Highway 42"},
                )
                print(result)  # Marvin, Ford, Unknown
            ```
        """  # noqa
        inputs = dict(
            command=operation,
            params=parameters,
            **execute_kwargs,
        )
        new, cursor = self._get_cursor(inputs)
        if new:
            await self._execute_async(cursor, inputs)
        self.logger.debug("Preparing to fetch all rows.")
        result = await run_sync_in_worker_thread(cursor.fetchall)
        return result

    @sync_compatible
    async def execute(
        self,
        operation: str,
        parameters: Optional[Dict[str, Any]] = None,
        **execute_kwargs: Dict[str, Any],
    ) -> None:
        """
        Executes an operation on the database. This method is intended to be used
        for operations that do not return data, such as INSERT, UPDATE, or DELETE.
        Unlike the fetch methods, this method will always execute the operation
        upon calling.

        Args:
            operation: The SQL query or other operation to be executed.
            parameters: The parameters for the operation.
            **execute_kwargs: Additional options to pass to `cursor.execute_async`.

        Examples:
            Create table named customers with two columns, name and address.
            ```python
            from prefect_snowflake.database import SnowflakeConnector

            with SnowflakeConnector.load("BLOCK_NAME") as conn:
                conn.execute(
                    "CREATE TABLE IF NOT EXISTS customers (name varchar, address varchar);"
                )
            ```
        """  # noqa
        self._start_connection()

        inputs = dict(
            command=operation,
            params=parameters,
            **execute_kwargs,
        )
        with self._connection.cursor() as cursor:
            await run_sync_in_worker_thread(cursor.execute, **inputs)
        self.logger.info(f"Executed the operation, {operation!r}.")

    @sync_compatible
    async def execute_many(
        self,
        operation: str,
        seq_of_parameters: List[Dict[str, Any]],
    ) -> None:
        """
        Executes many operations on the database. This method is intended to be used
        for operations that do not return data, such as INSERT, UPDATE, or DELETE.
        Unlike the fetch methods, this method will always execute the operations
        upon calling.

        Args:
            operation: The SQL query or other operation to be executed.
            seq_of_parameters: The sequence of parameters for the operation.

        Examples:
            Create table and insert three rows into it.
            ```python
            from prefect_snowflake.database import SnowflakeConnector

            with SnowflakeConnector.load("BLOCK_NAME") as conn:
                conn.execute(
                    "CREATE TABLE IF NOT EXISTS customers (name varchar, address varchar);"
                )
                conn.execute_many(
                    "INSERT INTO customers (name, address) VALUES (%(name)s, %(address)s);",
                    seq_of_parameters=[
                        {"name": "Marvin", "address": "Highway 42"},
                        {"name": "Ford", "address": "Highway 42"},
                        {"name": "Unknown", "address": "Space"},
                    ],
                )
            ```
        """  # noqa
        self._start_connection()

        inputs = dict(
            command=operation,
            seqparams=seq_of_parameters,
        )
        with self._connection.cursor() as cursor:
            await run_sync_in_worker_thread(cursor.executemany, **inputs)
        self.logger.info(
            f"Executed {len(seq_of_parameters)} operations off {operation!r}."
        )

    def close(self):
        """
        Closes connection and its cursors.
        """
        try:
            self.reset_cursors()
        finally:
            if self._connection is None:
                self.logger.info("There was no connection open to be closed.")
                return
            self._connection.close()
            self._connection = None
            self.logger.info("Successfully closed the Snowflake connection.")

    def __enter__(self):
        """
        Start a connection upon entry.
        """
        return self

    def __exit__(self, *args):
        """
        Closes connection and its cursors upon exit.
        """
        self.close()

    def __getstate__(self):
        """Allows block to be pickled and dumped."""
        data = self.__dict__.copy()
        data.update({k: None for k in {"_connection", "_unique_cursors"}})
        return data

    def __setstate__(self, data: dict):
        """Reset connection and cursors upon loading."""
        self.__dict__.update(data)
        self._start_connection()


@task
async def snowflake_query(
    query: str,
    snowflake_connector: SnowflakeConnector,
    params: Union[Tuple[Any], Dict[str, Any]] = None,
    cursor_type: SnowflakeCursor = SnowflakeCursor,
    poll_frequency_seconds: int = 1,
) -> List[Tuple[Any]]:
    """
    Executes a query against a Snowflake database.

    Args:
        query: The query to execute against the database.
        params: The params to replace the placeholders in the query.
        snowflake_connector: The credentials to use to authenticate.
        cursor_type: The type of database cursor to use for the query.
        poll_frequency_seconds: Number of seconds to wait in between checks for
            run completion.

    Returns:
        The output of `response.fetchall()`.

    Examples:
        Query Snowflake table with the ID value parameterized.
        ```python
        from prefect import flow
        from prefect_snowflake.credentials import SnowflakeCredentials
        from prefect_snowflake.database import SnowflakeConnector, snowflake_query


        @flow
        def snowflake_query_flow():
            snowflake_credentials = SnowflakeCredentials(
                account="account",
                user="user",
                password="password",
            )
            snowflake_connector = SnowflakeConnector(
                database="database",
                warehouse="warehouse",
                schema="schema",
                credentials=snowflake_credentials
            )
            result = snowflake_query(
                "SELECT * FROM table WHERE id=%{id_param}s LIMIT 8;",
                snowflake_connector,
                params={"id_param": 1}
            )
            return result

        snowflake_query_flow()
        ```
    """
    # context manager automatically rolls back failed transactions and closes
    with snowflake_connector.get_connection() as connection:
        with connection.cursor(cursor_type) as cursor:
            response = cursor.execute_async(query, params=params)
            query_id = response["queryId"]
            while connection.is_still_running(
                connection.get_query_status_throw_if_error(query_id)
            ):
                await asyncio.sleep(poll_frequency_seconds)
            cursor.get_results_from_sfqid(query_id)
            result = cursor.fetchall()
    return result


@task
async def snowflake_multiquery(
    queries: List[str],
    snowflake_connector: SnowflakeConnector,
    params: Union[Tuple[Any], Dict[str, Any]] = None,
    cursor_type: SnowflakeCursor = SnowflakeCursor,
    as_transaction: bool = False,
    return_transaction_control_results: bool = False,
    poll_frequency_seconds: int = 1,
) -> List[List[Tuple[Any]]]:
    """
    Executes multiple queries against a Snowflake database in a shared session.
    Allows execution in a transaction.

    Args:
        queries: The list of queries to execute against the database.
        params: The params to replace the placeholders in the query.
        snowflake_connector: The credentials to use to authenticate.
        cursor_type: The type of database cursor to use for the query.
        as_transaction: If True, queries are executed in a transaction.
        return_transaction_control_results: Determines if the results of queries
            controlling the transaction (BEGIN/COMMIT) should be returned.
        poll_frequency_seconds: Number of seconds to wait in between checks for
            run completion.

    Returns:
        List of the outputs of `response.fetchall()` for each query.

    Examples:
        Query Snowflake table with the ID value parameterized.
        ```python
        from prefect import flow
        from prefect_snowflake.credentials import SnowflakeCredentials
        from prefect_snowflake.database import SnowflakeConnector, snowflake_multiquery


        @flow
        def snowflake_multiquery_flow():
            snowflake_credentials = SnowflakeCredentials(
                account="account",
                user="user",
                password="password",
            )
            snowflake_connector = SnowflakeConnector(
                database="database",
                warehouse="warehouse",
                schema="schema",
                credentials=snowflake_credentials
            )
            result = snowflake_multiquery(
                ["SELECT * FROM table WHERE id=%{id_param}s LIMIT 8;", "SELECT 1,2"],
                snowflake_connector,
                params={"id_param": 1},
                as_transaction=True
            )
            return result

        snowflake_multiquery_flow()
        ```
    """
    with snowflake_connector.get_connection() as connection:
        if as_transaction:
            queries.insert(0, BEGIN_TRANSACTION_STATEMENT)
            queries.append(END_TRANSACTION_STATEMENT)

        with connection.cursor(cursor_type) as cursor:
            results = []
            for query in queries:
                response = cursor.execute_async(query, params=params)
                query_id = response["queryId"]
                while connection.is_still_running(
                    connection.get_query_status_throw_if_error(query_id)
                ):
                    await asyncio.sleep(poll_frequency_seconds)
                cursor.get_results_from_sfqid(query_id)
                result = cursor.fetchall()
                results.append(result)

    # cut off results from BEGIN/COMMIT queries
    if as_transaction and not return_transaction_control_results:
        return results[1:-1]
    else:
        return results


@task
async def snowflake_query_sync(
    query: str,
    snowflake_connector: SnowflakeConnector,
    params: Union[Tuple[Any], Dict[str, Any]] = None,
    cursor_type: SnowflakeCursor = SnowflakeCursor,
) -> List[Tuple[Any]]:
    """
    Executes a query in sync mode against a Snowflake database.

    Args:
        query: The query to execute against the database.
        params: The params to replace the placeholders in the query.
        snowflake_connector: The credentials to use to authenticate.
        cursor_type: The type of database cursor to use for the query.

    Returns:
        The output of `response.fetchall()`.

    Examples:
        Execute a put statement.
        ```python
        from prefect import flow
        from prefect_snowflake.credentials import SnowflakeCredentials
        from prefect_snowflake.database import SnowflakeConnector, snowflake_query


        @flow
        def snowflake_query_sync_flow():
            snowflake_credentials = SnowflakeCredentials(
                account="account",
                user="user",
                password="password",
            )
            snowflake_connector = SnowflakeConnector(
                database="database",
                warehouse="warehouse",
                schema="schema",
                credentials=snowflake_credentials
            )
            result = snowflake_query_sync(
                "put file://afile.csv @mystage;",
                snowflake_connector,
            )
            return result

        snowflake_query_sync_flow()
        ```
    """
    # context manager automatically rolls back failed transactions and closes
    with snowflake_connector.get_connection() as connection:
        with connection.cursor(cursor_type) as cursor:
            cursor.execute(query, params=params)
            result = cursor.fetchall()
    return result



"""Credentials block for authenticating with Snowflake."""

import re
import warnings
from pathlib import Path
from typing import Any, Dict, Optional, Union

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
    load_pem_private_key,
)

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import snowflake.connector
from prefect.blocks.abstract import CredentialsBlock
from pydantic import Field, SecretBytes, SecretField, SecretStr, root_validator

# PEM certificates have the pattern:
#   -----BEGIN PRIVATE KEY-----
#   <- multiple lines of encoded data->
#   -----END PRIVATE KEY-----
#
# The regex captures the header and footer into groups 1 and 3, the body into group 2
# group 1: "header" captures series of hyphens followed by anything that is
#           not a hyphen followed by another string of hyphens
# group 2: "body" capture everything upto the next hyphen
# group 3: "footer" duplicates group 1
_SIMPLE_PEM_CERTIFICATE_REGEX = "^(-+[^-]+-+)([^-]+)(-+[^-]+-+)"


class InvalidPemFormat(Exception):
    """Invalid PEM Format Certificate"""


class SnowflakeCredentials(CredentialsBlock):
    """
    Block used to manage authentication with Snowflake.

    Args:
        account (str): The snowflake account name.
        user (str): The user name used to authenticate.
        password (SecretStr): The password used to authenticate.
        private_key (SecretStr): The PEM used to authenticate.
        authenticator (str): The type of authenticator to use for initializing
            connection (oauth, externalbrowser, etc); refer to
            [Snowflake documentation](https://docs.snowflake.com/en/user-guide/python-connector-api.html#connect)
            for details, and note that `externalbrowser` will only
            work in an environment where a browser is available.
        token (SecretStr): The OAuth or JWT Token to provide when
            authenticator is set to OAuth.
        endpoint (str): The Okta endpoint to use when authenticator is
            set to `okta_endpoint`, e.g. `https://<okta_account_name>.okta.com`.
        role (str): The name of the default role to use.
        autocommit (bool): Whether to automatically commit.

    Example:
        Load stored Snowflake credentials:
        ```python
        from prefect_snowflake import SnowflakeCredentials

        snowflake_credentials_block = SnowflakeCredentials.load("BLOCK_NAME")
        ```
    """  # noqa E501

    _block_type_name = "Snowflake Credentials"
    _logo_url = "https://images.ctfassets.net/gm98wzqotmnx/2DxzAeTM9eHLDcRQx1FR34/f858a501cdff918d398b39365ec2150f/snowflake.png?h=250"  # noqa
    _documentation_url = "https://prefecthq.github.io/prefect-snowflake/credentials/#prefect_snowflake.credentials.SnowflakeCredentials"  # noqa

    account: str = Field(
        ..., description="The snowflake account name.", example="nh12345.us-east-2.aws"
    )
    user: str = Field(..., description="The user name used to authenticate.")
    password: Optional[SecretStr] = Field(
        default=None, description="The password used to authenticate."
    )
    private_key: Optional[SecretBytes] = Field(
        default=None, description="The PEM used to authenticate."
    )
    private_key_path: Optional[Path] = Field(
        default=None, description="The path to the private key."
    )
    private_key_passphrase: Optional[SecretStr] = Field(
        default=None, description="The password to use for the private key."
    )
    authenticator: Literal[
        "snowflake",
        "snowflake_jwt",
        "externalbrowser",
        "okta_endpoint",
        "oauth",
        "username_password_mfa",
    ] = Field(  # noqa
        default="snowflake",
        description=("The type of authenticator to use for initializing connection."),
    )
    token: Optional[SecretStr] = Field(
        default=None,
        description=(
            "The OAuth or JWT Token to provide when authenticator is set to `oauth`."
        ),
    )
    endpoint: Optional[str] = Field(
        default=None,
        description=(
            "The Okta endpoint to use when authenticator is set to `okta_endpoint`."
        ),
    )
    role: Optional[str] = Field(
        default=None, description="The name of the default role to use."
    )
    autocommit: Optional[bool] = Field(
        default=None, description="Whether to automatically commit."
    )

    @root_validator(pre=True)
    def _validate_auth_kwargs(cls, values):
        """
        Ensure an authorization value has been provided by the user.
        """
        auth_params = (
            "password",
            "private_key",
            "private_key_path",
            "authenticator",
            "token",
        )
        if not any(values.get(param) for param in auth_params):
            auth_str = ", ".join(auth_params)
            raise ValueError(
                f"One of the authentication keys must be provided: {auth_str}\n"
            )
        elif values.get("private_key") and values.get("private_key_path"):
            raise ValueError(
                "Do not provide both private_key and private_key_path; select one."
            )
        elif values.get("password") and values.get("private_key_passphrase"):
            raise ValueError(
                "Do not provide both password and private_key_passphrase; "
                "specify private_key_passphrase only instead."
            )
        return values

    @root_validator(pre=True)
    def _validate_token_kwargs(cls, values):
        """
        Ensure an authorization value has been provided by the user.
        """
        authenticator = values.get("authenticator")
        token = values.get("token")
        if authenticator == "oauth" and not token:
            raise ValueError(
                "If authenticator is set to `oauth`, `token` must be provided"
            )
        return values

    @root_validator(pre=True)
    def _validate_okta_kwargs(cls, values):
        """
        Ensure an authorization value has been provided by the user.
        """
        authenticator = values.get("authenticator")

        # did not want to make a breaking change so we will allow both
        # see https://github.com/PrefectHQ/prefect-snowflake/issues/44
        if "okta_endpoint" in values.keys():
            warnings.warn(
                "Please specify `endpoint` instead of `okta_endpoint`; "
                "`okta_endpoint` will be removed March 31, 2023.",
                DeprecationWarning,
                stacklevel=2,
            )
            # remove okta endpoint from fields
            okta_endpoint = values.pop("okta_endpoint")
            if "endpoint" not in values.keys():
                values["endpoint"] = okta_endpoint

        endpoint = values.get("endpoint")
        if authenticator == "okta_endpoint" and not endpoint:
            raise ValueError(
                "If authenticator is set to `okta_endpoint`, "
                "`endpoint` must be provided"
            )
        return values

    def resolve_private_key(self) -> Optional[bytes]:
        """
        Converts a PEM encoded private key into a DER binary key.

        Returns:
            DER encoded key if private_key has been provided otherwise returns None.

        Raises:
            InvalidPemFormat: If private key is not in PEM format.
        """
        if self.private_key_path is None and self.private_key is None:
            return None
        elif self.private_key_path:
            private_key = self.private_key_path.read_bytes()
        else:
            private_key = self._decode_secret(self.private_key)

        if self.private_key_passphrase is not None:
            password = self._decode_secret(self.private_key_passphrase)
        elif self.password is not None:
            warnings.warn(
                "Using the password field for private_key is deprecated "
                "and will not work after March 31, 2023; please use "
                "private_key_passphrase instead",
                DeprecationWarning,
                stacklevel=2,
            )
            password = self._decode_secret(self.password)
        else:
            password = None

        composed_private_key = self._compose_pem(private_key)
        return load_pem_private_key(
            data=composed_private_key,
            password=password,
            backend=default_backend(),
        ).private_bytes(
            encoding=Encoding.DER,
            format=PrivateFormat.PKCS8,
            encryption_algorithm=NoEncryption(),
        )

    @staticmethod
    def _decode_secret(secret: Union[SecretStr, SecretBytes]) -> Optional[bytes]:
        """
        Decode the provided secret into bytes. If the secret is not a
        string or bytes, or it is whitespace, then return None.

        Args:
            secret: The value to decode.

        Returns:
            The decoded secret as bytes.

        """
        if isinstance(secret, (SecretBytes, SecretStr)):
            secret = secret.get_secret_value()

        if not isinstance(secret, (bytes, str)) or len(secret) == 0 or secret.isspace():
            return None

        return secret if isinstance(secret, bytes) else secret.encode()

    @staticmethod
    def _compose_pem(private_key: bytes) -> bytes:
        """Validate structure of PEM certificate.

        The original key passed from Prefect is sometimes malformed.
        This function recomposes the key into a valid key that will
        pass the serialization step when resolving the key to a DER.

        Args:
            private_key: A valid PEM format byte encoded string.

        Returns:
            byte encoded certificate.

        Raises:
            InvalidPemFormat: if private key is an invalid format.
        """
        pem_parts = re.match(_SIMPLE_PEM_CERTIFICATE_REGEX, private_key.decode())
        if pem_parts is None:
            raise InvalidPemFormat()

        body = "\n".join(re.split(r"\s+", pem_parts[2].strip()))
        # reassemble header+body+footer
        return f"{pem_parts[1]}\n{body}\n{pem_parts[3]}".encode()

    def get_client(
        self, **connect_kwargs: Dict[str, Any]
    ) -> snowflake.connector.SnowflakeConnection:
        """
        Returns an authenticated connection that can be used to query
        Snowflake databases.

        Any additional arguments passed to this method will be used to configure
        the SnowflakeConnection. For available parameters, please refer to the
        [Snowflake Python connector documentation](https://docs.snowflake.com/en/user-guide/python-connector-api.html#connect).

        Args:
            **connect_kwargs: Additional arguments to pass to
                `snowflake.connector.connect`.

        Returns:
            An authenticated Snowflake connection.

        Example:
            Get Snowflake connection with only block configuration:
            ```python
            from prefect_snowflake import SnowflakeCredentials

            snowflake_credentials_block = SnowflakeCredentials.load("BLOCK_NAME")

            connection = snowflake_credentials_block.get_client()
            ```

            Get Snowflake connector scoped to a specified database:
            ```python
            from prefect_snowflake import SnowflakeCredentials

            snowflake_credentials_block = SnowflakeCredentials.load("BLOCK_NAME")

            connection = snowflake_credentials_block.get_client(database="my_database")
            ```
        """  # noqa
        connect_params = {
            **connect_kwargs,
            # required to track task's usage in the Snowflake Partner Network Portal
            "application": "Prefect_Snowflake_Collection",
            **self.dict(exclude_unset=True, exclude={"block_type_slug"}),
        }

        for key, value in connect_params.items():
            if isinstance(value, SecretField):
                connect_params[key] = connect_params[key].get_secret_value()

        # set authenticator to the actual okta_endpoint
        if connect_params.get("authenticator") == "okta_endpoint":
            endpoint = connect_params.pop("endpoint", None) or connect_params.pop(
                "okta_endpoint", None
            )  # okta_endpoint is deprecated
            connect_params["authenticator"] = endpoint

        private_der_key = self.resolve_private_key()
        if private_der_key is not None:
            connect_params["private_key"] = private_der_key
            connect_params.pop("password", None)
            connect_params.pop("private_key_passphrase", None)

        return snowflake.connector.connect(**connect_params)
